# -*- coding: utf-8 -*-
#
# This program tests the generalised Schur algorithm to find the 
# Cholesky decomposition of the inverse covariance matrix.
#
#  This script is part of HectorP 0.1.1
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 21/2/2021 Machiel Bos, Santa Clara
#===============================================================================

import os
import math
import sys
import re
import argparse
import numpy as np
from hectorp.powerlaw import Powerlaw

#===============================================================================
# Subroutines
#===============================================================================

def levinson(t):
    ''' Use Durbin-Levinson algorithm to compute l1 and l2

    Args:
        t : array containing first column of Toeplitz matrix

    Returns:
        l1,l2 : arrays containing Schur polynomials
        delta : scale factor
    '''

    #--- Durbin-Levinson to compute l1 and l2
    r = np.zeros(m-1)
    delta = t[0]
    for i in range(0,m-1):
        if i==0:
            gamma = -t[i+1]/delta
        else:
            gamma = -(t[i+1] + np.dot(t[1:i+1],r[0:i]))/delta
            r[1:i+1] = r[0:i] + gamma*r[i-1::-1]

        r[0] = gamma
        delta = t[0] + np.dot(t[1:i+2],r[i::-1])

    #--- create l1 & l2 using r
    l1 = np.zeros(2*m)
    l2 = np.zeros(2*m)

    l1[0]   = 1.0
    l1[1:m] = r[m-2::-1]
    l2[1:m] = r[0:m-1]

    return [l1,l2,delta]



#===============================================================================
# Main program
#===============================================================================

def main():

    noise_model = Powerlaw()
    t = noise_model.create_t(4,0,[-0.8])
    print(t)



#!/usr/bin/env python3
# 
# Apply wiener filter to estimate the varying seasonal signal
#
# INPUT
# -----
# Station name : example CASC
# sigma_a      : std of random variation of annual amplitude
# sigma_sa     : std of random variation of semi-annual amplitude
# phi          : coefficient AR(1) process
#
# OUTPUT
# ------
# file with estimated seasonal signal
# file with observations - seasonal signal
#
# Machiel Bos, 30/4/2018, Coimbra
#
#  This script is part of Hector 2.1
#
#  Hector is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  Hector is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with Hector.  If not, see <http://www.gnu.org/licenses/>
#===============================================================================

import sys
import os
import math
import json
import numpy as np
import subprocess

#===============================================================================
# Global constants
#===============================================================================

pi  = 4.0*math.atan(1.0)
EPS = 1.0e-8


#===============================================================================
# Functions
#===============================================================================

#-----------------------------------------
def model_PSD_S(omega,phi,sigma_s,omega0):
#-----------------------------------------
    """ Computes the power spectral density of the periodic signal.

    Parameters
    ----------
    omega :
        normalised angular velocity [rad]
    phi :
        coefficients of AR(1) process
    sigma_s :
        standard deviation of white noise that drives AR(1) [mm]
    omega0 :
        normalised angular velocity of periodic signal [rad]

    Returns
    -------
    float
        power spectral density [mm^2/rad]
    """ 
    return 2.0*(pow(sigma_s,2.0)/pi) * \
		(1.0/(1.0-2.0*phi*math.cos(omega+omega0)+pow(phi,2.0)) + \
                 1.0/(1.0-2.0*phi*math.cos(omega-omega0)+pow(phi,2.0)))



#---------------------------------------------
def model_PSD_W(omega,kappa,sigma_pl,sigma_w):
#---------------------------------------------
    """ Computes the power spectral density of the noise.

    Parameters
    ----------
    omega :
        normalised angular velocity [rad]
    kappa :
        spectral index
    sigma_pl :
        standard deviation of the power-law noise [mm]
    sigma_w :
        standard deviation of the white noise [mm]

    Returns
    -------
    float
        power spectral density [mm^2/rad]
    """ 
    if omega<1.0e-6:
        return 9.9e99
    else:
        return (1.0/pi) * (pow(sigma_pl,2.0)/ \
		pow(2.0*math.sin(0.5*omega),-kappa) + pow(sigma_w,2.0));



#-----------------------------------------------------------------
def wienerfilter(n,x,kappa,sigma_pl,sigma_w,sigma_a,sigma_sa,phi):
#-----------------------------------------------------------------
    """ Applies the Wiener Filter to the residuals in vector r

    Parameters
    ----------
    n :
        number of observations
    x :
        vector with residuals
    kappa :
        spectral index
    sigma_pl :
        standard deviation of the power-law noise [mm]
    sigma_w :
        standard deviation of the white noise [mm]
    sigma_a :
        standard deviation of white noise that drives annual AR(1) [mm]
    sigma_sa :
        standard deviation of white noise that drives semi-annual AR(1) [mm]
    phi :
        coefficients of AR(1) process

    Returns
    -------
    float
        vector with estimated varying seasonal signal (s_r) [mm]
    """
    #--- angular velocity of annual signal
    omega0 = 2*pi/365.25

    #--- Compute FFT of observed time series (column vector)
    xfft = np.fft.rfft(x.T,n).T

    for i in range(0,int(n/2)+1):

        #--- frequency in cylcles per day
        omega = 2*pi*i/n

        #--- Compute scaling of FFT 
        S  = model_PSD_S(omega,phi,sigma_a,omega0)     # annual signal
        S += model_PSD_S(omega,phi,sigma_sa,2*omega0)  # semi-annual signal
        W  = model_PSD_W(omega,kappa,sigma_pl,sigma_w) # noise 
        H  = S/(S+W)                                   # optimal filter

        #--- apply optimal filter
        xfft[i] *= H
 
    #--- Convert scaled FFT back to time domain (column vector)
    return np.squeeze(np.fft.irfft(xfft.T,n).T)



#===============================================================================
# Main program
#===============================================================================

#--- Constant
eps = 1.0e-6


#--- Read command line arguments
if len(sys.argv)!=5:
    print('Correct usage: apply_WF.py station_name sigma_a sigma_sa phi')
    sys.exit()
else:
    station_name = sys.argv[1]
    sigma_a      = float(sys.argv[2])
    sigma_sa     = float(sys.argv[3])
    phi          = float(sys.argv[4])


#--- Analyse mom file in directory ./obs_files
output = subprocess.check_output('analyse_timeseries.py -s {0:s} -n PLWN'. \
                                             format(station_name),shell=True)

#--- Check output before further parsing
if output.decode().startswith('Cannot')==True:
    print(output)
    sys.exit()


#--- parse output
fp_dummy = open('estimatetrend.json','r')
results = json.load(fp_dummy)
fp_dummy.close()

cos_annual        = results['cos_365.250']
sin_annual        = results['sin_365.250']
cos_annual_sigma  = results['cos_365.250_sigma']
sin_annual_sigma  = results['sin_365.250_sigma']
cos_hannual       = results['cos_182.625']
sin_hannual       = results['sin_182.625']
cos_hannual_sigma = results['cos_182.625_sigma']
sin_hannual_sigma = results['sin_182.625_sigma']
noise_model       = results['NoiseModel']
noise             = noise_model['White']
sigma_w           = noise['sigma']
noise             = noise_model['GGM']
sigma_pl          = noise['sigma']
d                 = noise['d']
                                                   
print('sigma_pl={0:f}'.format(sigma_pl))
print('sigma_w ={0:f}'.format(sigma_w))
print('d={0:f}'.format(d))

#--- Read file
header = ""
DeltaT = None
flag   = []   # True for observation, False for data gap
t      = []   # time in MJD
x      = []   # Observations
x_hat  = []   # Fitted model
mjd0   = None
with open('./mom_files/{0:s}.mom'.format(station_name),'r') as fp:
    line = fp.readline()
    #--- Read header & offsets
    while line[0]=="#":
        #--- add line to header
        header += line

        #--- Does it contain informatioon about the sampling period?
        if line[0:17]=='# sampling period':
            cols = line.split()
            DeltaT = float(cols[3])

        #--- Look at next line  
        line = fp.readline()
 
    #--- Remove last newline
    header = header.rstrip()
 
    #--- Read data
    while len(line)>0:
        cols = line.split()
        mjd  = float(cols[0])
        obs  = float(cols[1])
        mod  = float(cols[2])

        #--- Store the residual if it's the first or next in line
        if mjd0==None or abs(mjd-mjd0-DeltaT)<eps:
            t.append(float(mjd))
            x.append(float(obs))
            x_hat.append(float(mod))
            flag.append(True)
        else:
            #--- Fill missing data with zeros
            while mjd0+DeltaT+eps<mjd:
                mjd0 += DeltaT
                t.append(float(mjd0))
                x.append(float(0.0))
                x_hat.append(float(0.0))
                flag.append(False)
            t.append(float(mjd))
            x.append(float(obs))
            x_hat.append(float(mod))
            flag.append(True)
        
        #--- Shift mjd0 to current epoch (i.e. mjd)      
        mjd0 = mjd

        #--- Read next line
        line = fp.readline()

#--- Determine the number of observations (including data gaps)
n = len(t)

#--- Create residuals
shape = (n, 1)
r = np.zeros(shape)
for i in range(0,n):
    r[i] = x[i] - x_hat[i]

#--- Convert sigma_pl to my unit which is without (Delta T)^(-kappa/4)
kappa = -2.0*float(d)
if DeltaT==None:
    print('Assuming default sampling period of 1 day')
    DeltaT=1.0
sigma_pl *= pow(DeltaT/365.25,-kappa/4.0)


#--- Define Design matrix for seasonal signal (annual + semi-annual)
shape = (n, 4)
H = np.zeros(shape)
for i in range(0,n):
    H[i,0] = math.cos(2*pi*(t[i]-51544.0)/365.25);
    H[i,1] = math.sin(2*pi*(t[i]-51544.0)/365.25);
    H[i,2] = math.cos(4*pi*(t[i]-51544.0)/365.25);
    H[i,3] = math.sin(4*pi*(t[i]-51544.0)/365.25);

#--- Fill amplitudes of seasonal signal
shape = (4, 1)
theta = np.zeros(shape)
theta[0] = cos_annual
theta[1] = sin_annual
theta[2] = cos_hannual
theta[3] = sin_hannual

#--- Remember constant seasonal signal (annual + semi-annual)
s_c = np.squeeze(np.asarray(H.dot(theta)))

#for i in range(0,n):
#    print('{0:f},  {1:f},  {2:f}'.format(t[i],r[i,0],s_c[i]))

#--- Apply filter
s_r = wienerfilter(n,r,kappa,sigma_pl,sigma_w,sigma_a,sigma_sa,phi)



#----------------------
#--- Save results -----
#----------------------

#--- Does the sea_files directory exists?
if not os.path.exists('./sea_files'):
    os.mkdir('./sea_files/')

#--- Does the fil_files directory exists?
if not os.path.exists('./fil_files'):
    os.mkdir('./fil_files/')

fp_fil = open('./fil_files/{0:s}.mom'.format(station_name),'w')
fp_sea = open('./sea_files/{0:s}.mom'.format(station_name),'w')
fp_mom = open('./mom_files/{0:s}_WF.mom'.format(station_name),'w')

#--- Copy header
fp_fil.write('{0:s}\n'.format(header))
fp_sea.write('{0:s}\n'.format(header))
fp_mom.write('{0:s}\n'.format(header))

#--- Save filtered observations and estimated varying seasonal
for i in range(0,n):
    if flag[i]==True:
        fp_fil.write('{0:10.1f} {1:9.5f}\n'.format(t[i],x[i]-(s_c[i]+s_r[i])))
        fp_sea.write('{0:10.1f} {1:9.5f}\n'.format(t[i],s_c[i]+s_r[i]))
        fp_mom.write('{0:10.1f} {1:9.5f} {2:9.5f}\n'.format(t[i],x[i],x_hat[i]+s_r[i]))
fp_fil.close()
fp_sea.close()
fp_mom.close()



# covariance.py
#
# Create the first row of the covariance matrix for various noise models:
#  1) power-law noise using Eq. (7) of Bos et al. (2008).
#  2) white noise
#
# Bos, MS, Fernandes, RMS, Williams, SDP & Bastos, L (2008). "Fast error 
# analysis of continuous GPS observations". Journal of Geodesy, 82(3), 157-166.
#
#
# This file is part of HectorP 0.1.1.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 18/ 5/2020 Machiel Bos, Santa Clara
# 29/12/2021 Machiel Bos, Santa Clara
#==============================================================================

import numpy as np
import sys
import math
from hectorp.white import White
from hectorp.ggm import GGM
from hectorp.powerlaw import Powerlaw
from hectorp.varyingannual import VaryingAnnual
from hectorp.ar1 import AR1
from hectorp.matern import Matern
from hectorp.control import Control
from hectorp.control import SingletonMeta

#==============================================================================
# Subroutines
#==============================================================================


class Covariance(metaclass=SingletonMeta):


    def __init__(self):
        """ initialise class
        """

        #--- Get control parameters
        control = Control()
        try:
            self.verbose = control.params['Verbose']
        except:
            self.verbose = True

        try:
            self.noisemodel_names = control.params['NoiseModels']
            if isinstance(self.noisemodel_names,list)==False:
                self.noisemodel_names = [self.noisemodel_names]
        except Exception as e:
            print(e)
            sys.exit()

        #--- Create list of noise model class instances
        self.noisemodels = []
        i = 0
        for noisemodel_name in self.noisemodel_names:
            if noisemodel_name in ['FlickerGGM','RandomWalkGGM','GGM']:
                ClassName = getattr(sys.modules[__name__], 'GGM')
                if noisemodel_name=='FlickerGGM':
                    class_ = ClassName(0.5)
                elif noisemodel_name=='RandomWalkGGM':
                    class_ = ClassName(1.0)
                else: 
                    try:
                        d = -0.5*control.params['kappa_fixed']
                    except:
                        d = math.nan
                    class_ = ClassName(d)
            elif noisemodel_name=='Matern':
                try:
                    d = -0.5*control.params['kappa_fixed']
                except:
                    d = math.nan
                ClassName = getattr(sys.modules[__name__], noisemodel_name)
                class_ = ClassName(d)

            else:
                ClassName = getattr(sys.modules[__name__], noisemodel_name)
                class_ = ClassName() 

            self.noisemodels.append(class_)
            if self.verbose==True:
                print('{0:d}) {1:s}'.format(i,noisemodel_name))
            i += 1

        self.Nmodels = len(self.noisemodels)
        self.Nparam = self.Nmodels-1  # weight parameters

        #--- Do we need to estimate additional noise parameters?
        for noisemodel in self.noisemodels:
            self.Nparam += noisemodel.get_Nparam()
        if self.verbose==True:
            print('Nparam : {0:d}'.format(self.Nparam))           


    def get_Nparam(self):
        """ Return the number of parameters to estimate numerically
        
        Returns:
            self.Nparam (int) : total number of parameters 
        """

        return self.Nparam


        
    def compute_fraction(self,i,param):
        """ Compute fraction of noise model i

        Args:
            i (int) : index of noise model
            param (array float) : parameters describing weight of noise models
        
        Returns:
            fraction (float) : fraction of noise model
        """

        #--- Constant
        hpi = 2.0*math.atan(1.0)

        #--- compute fraction
        if self.Nmodels==1:
            return 1.0
        else:
            fraction = 1.0
            for j in range(0,i):
                fraction *= math.sin(hpi*param[j]);
            if i<self.Nmodels-1:
                fraction *= math.cos(hpi*param[i]);

        #--- Some range checks
        if fraction>1.0:
            fraction = 1.0

        return pow(fraction,2.0)
   


    def compute_penalty(self,param):
        """ penalty for each fraction outside the [0:1] range

        Args:
            param (list float) : fractions of each noise model (Nmodels-1)

        Returns:
            penalty value (float)
        """

        #--- Constant
        LARGE = 1.0e8
 
        #--- first fractions
        penalty = 0.0
        for i in range(0,len(self.noisemodels)-1):
            if param[i]<0.0:
                penalty += (0.0-param[i])*LARGE
                param[i] = 0.0
            elif param[i]>1.0:
                penalty += (param[i]-1.0)*LARGE
                param[i] = 1.0

        #--- Extra penalties for noise model parameters
        k = len(self.noisemodels)-1
        for noisemodel in self.noisemodels:
            penalty += noisemodel.penalty(k,param)

        return penalty



    def create_t(self,m,param):
        """

        Args:
            m (int) : length of time series
            param (array float) : array of parameters to estimate

        Returns:
            t (array float) : first row of covariance matrix
        """

        #--- Create empty autocovariance array
        t = np.zeros(m)

        #--- Add autocovariance of each noise model
        k = self.Nmodels-1
        for i in range(0,self.Nmodels):
            fraction = self.compute_fraction(i,param)
            t_part,k_new = self.noisemodels[i].create_t(m,k,param)

            t += fraction * t_part
            k  = k_new

        return t



    def show_results(self,output,noise_params,sigma_eta):
        """ show estimated noiseparameters

        Args:
            output (dictionary) : where values for json file are saved
            noise_params (float-array) : fractions + noise model parameters
            sigma_eta (float) : driving noise value
        """

        if self.verbose==True:
            print('\nNoise Models\n------------')
        k = self.Nmodels-1
        output_block = {}
        for i in range(0,len(self.noisemodels)):
            noisemodel = self.noisemodels[i]
            output_single = {}
            fraction = self.compute_fraction(i,noise_params)

            if self.verbose==True:
                print('{0:s}:'.format(self.noisemodel_names[i]))
                print('fraction  = {0:7.5f}'.format(fraction))

            sigma = math.sqrt(fraction)*sigma_eta
            output_single['fraction'] = fraction
            #--- sigma is stored in the json inside each noise model class
            k = noisemodel.show_results(output_single,k,noise_params,sigma)

            output_block[self.noisemodel_names[i]] = output_single

        output['NoiseModel'] = output_block



# -*- coding: utf-8 -*-
#
# Perform cycles of offset detection for a single station [optionally 3D]
#  1) Compute noise and SLT model parameters
#  2) Compute for each epoch the new log-likelihood when an offset is added
#     but maintaining the noise parameters constant.
#
#  This script is part of HectorP 0.1.1
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 2/1/2022 Machiel Bos, Santa Clara
#===============================================================================

import os
import math
import time
import json
import argparse
import numpy as np
from hectorp.control import Control
from hectorp.control import SingletonMeta
from hectorp.observations import Observations
from hectorp.designmatrix import DesignMatrix
from hectorp.covariance import Covariance
from hectorp.mle import MLE

#===============================================================================
# Subroutines
#===============================================================================

def create_ctl_file(fname,station,noisemodels,useRMLE):
    """ Create findoffsets.ctl

    Args:
        fname (string) : name of ctl-file
        station (string) : name of station
        noisemodels (string) : PLWN, GGMWN, ...
        useRMLE (boolean) : use RMLE in analyses or not
    """

    #--- Create control.txt file for EstimateTrend
    fp = open(fname, "w")
    fp.write("DataFile            {0:s}.mom\n".format(station))
    fp.write("DataDirectory       ./raw_files\n")
    fp.write("OutputFile          ./obs_files/{0:s}.mom\n".format(station))
    fp.write("PhysicalUnit        mm\n")
    fp.write("ScaleFactor         1.0\n")
    fp.write("periodicsignals     365.25 182.625\n")
    fp.write("estimateoffsets     yes\n")
    if noisemodels == 'FNWN':
        fp.write("NoiseModels         FlickerGGM White\n")
    elif noisemodels == 'PLWN':
        fp.write("NoiseModels         GGM White\n")
    elif noisemodels == 'RWFNWN':
        fp.write("NoiseModels         RandomWalkGGM FlickerGGM White\n")
    elif noisemodels == 'WN':
        fp.write("NoiseModels         White\n")
    elif noisemodels == 'PL':
        fp.write("NoiseModels         GGM\n")
    elif noisemodels == 'FL':
        fp.write("NoiseModels         FlickerGGM\n")
    else:
        print("Unknown noise model: {0:s}".format(noisemodels))
        sys.exit()
    fp.write("GGM_1mphi           6.9e-06\n")
    if useRMLE==True:
        fp.write("useRMLE             yes\n")
    else:
        fp.write("useRMLE             no\n")
    fp.close()

        

#===============================================================================
# Main program
#===============================================================================

def main():

    print("\n***************************************")
    print("    findoffsets, version 0.1.1")
    print("***************************************")

    #--- Parse command line arguments in a bit more professional way
    parser = argparse.ArgumentParser(description= 'Find offsets in time series')

    #--- List arguments that can be given 
    parser.add_argument('-s', dest='station', action='store', required=True,
        default='findoffset.ctl', help="The name of the ctl-file")
    parser.add_argument('-t', dest='threshold', action='store', required=False,
        default='20.0', help="The MLE difference considered significant") 
    parser.add_argument('-threeD', action='store_true',
       required=False, help="process east, north and up at same time")
    parser.add_argument('-n', dest='noisemodels', action='store',default='PLWN',
       required=False, help="noisemodel combination (PLWN, FL, etc.)")
    parser.add_argument('-verbose', action='store_true', 
       required=False, help="Verbosity (save values each step in file)")
    parser.add_argument('-i', required=False, default='findoffset.ctl', \
                                      dest='fname', help='Name of control file')
    parser.add_argument('-useRMLE', action='store_true',
                                    required=False, help="use RMLE option")

    args = parser.parse_args()

    #--- parse command-line arguments
    use_3D      = args.threeD
    useRMLE     = args.useRMLE
    station     = args.station
    threshold   = float(args.threshold)
    noisemodels = args.noisemodels
    verbose     = args.verbose
    fname       = args.fname

    #--- If use_3D then already create correct filenames
    if use_3D==True:
        stations = []
        for i in range(3):
            stations.append('{0:s}_{1:d}.dat'.format(station,i))
    else:
        stations = [station]

    #--- Start the clock!
    start_time = time.time()

    #--- Create and read control parameters into dictionary (singleton class)
    create_ctl_file(fname,stations[0],noisemodels,useRMLE)
    control = Control(fname)

    #--- Just create Observations to get number of observations m
    observations = Observations()
    m            = observations.m

    #--- For each time series, compute new MLE for each possible offset and sum
    #    the results.
    j = 0
    new_offsets = []
    while True:

        #--- array to store log-likelihood differences (new - old) values
        dln_L_sum = np.zeros(m)

        for station in stations:

            #--- each time series, start with a clean instantiation of classes
            SingletonMeta.clear_all()
            create_ctl_file(fname,station,noisemodels,useRMLE)
            control = Control(fname)
            observations = Observations()
            for t in new_offsets:
                observations.add_offset(t) # add new offsets to class
            covariance   = Covariance()
            mle          = MLE()
  
            #--- Test for this time series the effect of adding offset at each t
            dln_L = mle.test_new_offset()

            #--- Show information on screen
            if verbose==True:
                max_value = max(dln_L)
                index     = dln_L.index(max_value)
                t         = observations.data.index[index] 
                print('==> {0:s} - best offset at {1:9.2f} : {2:9.3f}'.\
						format(station,t,max_value))

            #--- Simply sum delta log-likelihoods 
            dln_L_sum += dln_L 

        max_value = max(dln_L_sum)
        result = np.where(dln_L_sum==max_value)
        index = result[0][0]
        t = observations.data.index[index] 
        print('************************************************************')
        print('best offset at {0:9.2f} (index {1:d}) : {2:9.3f}'.
					format(t,index,max_value))
        print('************************************************************')

        if verbose==True:
            with open('findoffset_{0:d}.out'.format(j),'w') as fp:
                for i in range(0,m):
                    if not math.isnan(observations.data.iloc[i,0]):
                        fp.write('{0:11.5f} {1:10.5f}\n'.\
					format(observations.data.index[i],
								dln_L_sum[i]))

        #--- If significant, store the found offset
        if max_value>threshold:
            new_offsets.append(t)
        else:
            break

        #--- prepare next round
        j += 1


    #--- Save dictionary 'output' as json file
    output = {}
    output['offsets'] = new_offsets
    with open('findoffsets.json','w') as fp:
        json.dump(output, fp, indent=4)

    #--- Save result to obs_files directory
    fname_out = control.params['OutputFile']
    observations.write(fname_out)

    #--- Show time lapsed
    print("--- {0:8.3f} seconds ---\n".format(float(time.time() - start_time)))



# -*- coding: utf-8 -*-
#
# This small program computes the Modified Julian Date (MJD).
#
#  This script is part of HectorP 0.1.1
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 14/2/2022 Machiel Bos, Santa Clara
#===============================================================================

import sys
from hectorp.calendar import compute_mjd

#===============================================================================
# Main program
#===============================================================================

def main():

    args = sys.argv[1:]

    if len(args)!=6:
        print("Correct input: date2MJD year month day hour minute second")
        sys.exit()
    else:
        year  = int(args[0])
        month = int(args[1])
        day   = int(args[2])
        hour  = int(args[3])
        minute= int(args[4])
        second= float(args[5])
        mjd = compute_mjd(year,month,day,hour,minute,second)
        print("year   : {0:4d}".format(year))
        print("month  : {0:4d}".format(month))
        print("day    : {0:4d}".format(day))
        print("hour   : {0:4d}".format(hour))
        print("minute : {0:4d}".format(minute))
        print("second : {0:f}".format(second))
        print("MJD    : {0:f}".format(mjd))



# ggm.py
#
# Create the first row of the covariance matrix for Generalised Gauss Markov
# noise.
#
# This file is part of HectorP 0.1.1.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 12/1/2022 Machiel Bos, Santa Clara
#==============================================================================

import numpy as np
import sys
import math
from mpmath import *
from hectorp.control import Control
from hectorp.observations import Observations

#==============================================================================
# Subroutines
#==============================================================================


class GGM:

    def __init__(self,d_fixed=math.nan):
        """ initialise class
        """

        #--- Set precision mpmath
        mp.dps = 25

        #--- Get control parameters
        control = Control()

        #--- Remember the d value used to instantiate this class
        self.d_fixed = d_fixed

        #--- Check if 1-phi is given in control file
        try:
            self.phi_fixed = control.params['GGM_1mphi']
        except:
            self.phi_fixed = math.nan

        #--- Number of noise parameters
        self.Nparam = 2
        if not math.isnan(self.d_fixed):
            self.Nparam -= 1
        if not math.isnan(self.phi_fixed):
            self.Nparam -= 1



    def get_Nparam(self):
        """ Return the number of parameters in GGM noise model
        
        Returns:
            self.Nparam (int) : total number of parameters (free)
        """

        return self.Nparam



    def backward(self,a,b,c,z,F,Fp1):
        """ Compute backward recursion

        Args:
            a,b,c,z (double) : Hypergeometric function 2F1(a,b;c;z)
            Fp1 (double)     : 2F1(a+1,b;c+1;z)

        Returns:
            2F1(a-1,b;c-1;z)
        """

        return ((1.0-c+(b-a)*z)*F + (a*(c-b)*z)*Fp1/c)/(1.0-c)


    
    def create_t(self,m,k,param):
        """ Create first row of covariance matrix of power-law noise
    
        Args:
            m (int) : length of time series
            k (int) : index of param
            param (array float) : spectral index
        
        Returns:
            t (row (m,1)) : first row Toeplitz covariance matrix 
            k_new (int)   : shifted index in param array
        """

        #--- Constant
        EPS = 1.0e-12

        #--- extract parameters to readable variables
        if self.Nparam==0:
            d     = self.d_fixed
            kappa = -2.0*d
            phi   = self.phi_fixed
            k_new = k
        elif self.Nparam==1:
            kappa = param[k]
            d     = -0.5*kappa
            phi   = self.phi_fixed
            k_new = k+1   # increase k for next model
        else:
            kappa = param[k+0]
            d     = -0.5*kappa
            phi   = param[k+1]
            k_new = k+2   # increase k for next model

        #--- Create first row vector of Covariance matrix
        t = np.zeros(m)

        #--- Create array with hypergeometric 2F1 values
        _2F1 = np.zeros(m)
       
        #--- for phi=0, we have pure power-law noise
        if fabs(phi)<EPS:
            if d>0.5:
                print("kappa< -1.0 ({0:f}) : non-stationary".format(kappa))
                print("1-phi: {0:f}".format(phi))
                sys.exit()
       
            #--- compute power-law noise 
            t[0] = math.gamma(1.0+kappa)/pow(math.gamma(1+0.5*kappa),2.0) 
            for i in range(1,m):
                t[i] = (i - 0.5*kappa - 1.0)/(i + 0.5*kappa) * t[i-1]

        #--- Not pure power-law noise
        else:
            #--- For d=0, _2F1 is always 1.0
            if fabs(d)<EPS:
                for i in range(0,m):
                    _2F1[i] = 1.0
            else:
                #--- Since phi is actually stored as 1-phi, I here need to
                #    put 1- (1-phi) = phi. DONT DELETE THIS COMMENT!!!
                z = math.pow(1-phi,2.0)
                k = m-1
                b = d
                a = d   + float(k)
                c = 1.0 + float(k)
                _2F1[m-1]= hyp2f1(a,b, c, z)
                a -= 1.0
                c -= 1.0
                _2F1[m-2]= hyp2f1(a,b, c, z)

                Fp1 = _2F1[m-1]
                F   = _2F1[m-2]
                for i in range(m-3,-1,-1):
                    _2F1[i] = self.backward(a,b,c,z,F,Fp1)
                    Fm1 = _2F1[i]

                    #--- prepare next round
                    a  -= 1.0
                    c  -= 1.0
                    Fp1 = F
                    F   = Fm1

        #--- finally, construct gamma_x
        scale = 1.0;
        for i in range(0,m):
            t[i]   = scale*_2F1[i]
            scale *= (d+float(i))*(1.0-phi)/(float(i)+1.0)
            if math.isnan(t[i]):
                print("Trouble in paradise!")
                print("i={0:d}, d={1:f}, 1-phi={2:e}".format(i,d,phi))
                sys.exit()
        
        return t, k_new 



    def penalty(self,k,param):
        """ Computes penalty for power-law noise

        Args:
            k (int) : index of param
            param (array float) : spectral index
        
        Returns:
            penalty (float)
        """

        LARGE = 1.0e8
        penalty = 0.0 

        if self.Nparam==0:
            penalty = 0.0 
        elif self.Nparam==1:
            kappa = param[k]
            if kappa<-3.0:
                penalty = (3.0 - kappa)*LARGE
                param[k] = -3.0
            elif kappa>0.01:
                penalty = (kappa - 0.01)*LARGE
                param[k] = 0.01
        else:
            d   = -0.5 * param[k]
            phi = param[k+1]

            #--- Extra checks to avoid danger zone
            if phi>0.0:
                y = math.log10(phi)
            else:
                y = 9.9e99         # will not be used

            #--- Check if log(1-phi) is below the line (2F1 is too large)
            safety_factor = 2.0
            if y < (4.0*d - 11.0 + safety_factor):
                penalty = ((4.0*d-11.0+safety_factor) - y)*LARGE
                param[k+1] = pow(10,4.0*d-11.0+safety_factor)

            #--- param[k+1] is always 1-phi. The following rarely occurs
            elif phi>0.999:
                penalty = (phi-0.999)*LARGE
                param[k+1] = 0.999

            #--- The following limit is most critical, stay away from zero!!
            #    Another complication is that at phi=0, you have power-law and
            #    then d_max=0.5. Thus, a jump down from 2. Allowing this is 
            #    asking for trouble. I put lower limit to 1.0e-6.
            elif phi<1.0e-6:
                penalty = (1.0e-6-phi)*LARGE
                param[k+1] = 1.0e-6
                
        return penalty



    def show_results(self,output_single,k,noise_params,sigma):
        """ show estimated noiseparameters

        Args:
            output_single (dictionary) : where values for json file are saved
            k (int) : index where we should start reading noise_params
            noise_params (float-array) : fractions + noise model parameters
            sigma (float) : noise amplitude of power-law noise
        """
      
        #--- Get some info from other classes
        control = Control() 
        observations = Observations()
        unit = control.params['PhysicalUnit']
        try:
            verbose = control.params['Verbose']
        except:
            verbose = True

        if observations.ts_format=='mom':
            T = observations.sampling_period/365.25 # T fraction -> year
            timeunit = 'yr'
        elif observations.ts_format=='msf':
            T = observations.sampling_period/3600.0 # T fraction -> hour
            timeunit = 'h'
        else:
            print('unknown ts_format {0:s}'.format(observations.ts_format))
            sys.exit()

        if self.Nparam==0:
            d = self.d_fixed
            kappa = -2.0*d
            phi = self.phi_fixed
        else:
            kappa = noise_params[k]
            d     = -0.5*kappa
            if self.Nparam==1:
                phi = self.phi_fixed
            else:
                phi = noise_params[k+1]

        sigma /= math.pow(T,0.5*d)
        if verbose==True:
            print('sigma     = {0:7.4f} {1:s}/{2:s}^{3:.2f}'.format(sigma,
								unit,timeunit,0.5*d))

            if self.Nparam==0:
                print('d         = {0:7.4f} (fixed)'.format(d))
                print('kappa     = {0:7.4f} (fixed)'.format(kappa))
                print('1-phi     = {0:7.4f} (fixed)\n'.format(phi))
            elif self.Nparam==1:
                print('d         = {0:7.4f}'.format(d))
                print('kappa     = {0:7.4f}'.format(kappa))
                print('1-phi     = {0:7.4f} (fixed)\n'.format(phi))
            else:
                print('d         = {0:7.4f}'.format(d))
                print('kappa     = {0:7.4f}'.format(kappa))
                print('1-phi     = {0:7.4f}\n'.format(phi))

        output_single['d']     = d
        output_single['kappa'] = kappa
        output_single['1-phi'] = phi
        output_single['sigma'] = sigma

        return k+self.Nparam



# observations.py
#
# A simple interface that reads and writes mom-files and stores
# them into a Python class 'observations'.
#
# This file is part of HectorP 0.1.1.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
#  31/1/2019  Machiel Bos, Santa Clara
#   1/5/2020  David Bugalho
# 29/12/2021  Machiel Bos, Santa Clara
#  7/ 2/2022  Machiel Bos, Santa Clara
# 10/ 7/2022  Machiel Bos, Santa Clara
#==============================================================================

import pandas as pd
import numpy as np
import os
import sys
import math
import re
from hectorp.control import Control
from hectorp.control import SingletonMeta
from pathlib import Path

#==============================================================================
# Class definition
#==============================================================================

class Observations(metaclass=SingletonMeta):
    """Class to store my time series together with some metadata
    
    Methods
    -------
    momread(fname)
        read mom-file fname and store the data into the mom class
    momwrite(fname)
        write the momdata to a file called fname
    msfread(fname)
        read msf-file fname and store the data into the mom class
    msfwrite(fname)
        write the msfdata to a file called fname
    make_continuous()
        make index regularly spaced + fill gaps with NaN's
    """
    
    def __init__(self):
        """This is my time series class
        
        This constructor defines the time series in pandas DataFrame data,
        list of offsets and the sampling period (unit days)
        
        """

        #--- Get control parameters (singleton)
        control = Control()
        try:
            self.verbose = control.params['Verbose']
        except:
            self.verbose = True

        #--- Scale factor
        try:
            self.scale_factor = float(control.params['ScaleFactor'])
        except:
            self.scale_factor = 1.0

        #--- class variables
        self.data = pd.DataFrame()
        self.offsets = []
        self.postseismicexp = []
        self.postseismiclog = []
        self.ssetanh = []
        self.sampling_period = 0.0
        self.F = None
        self.percentage_gaps = None
        self.m = 0
        self.column_name=''

        #--- Read filename with observations and the directory
        try:
            self.datafile = control.params['DataFile']
            directory = Path(control.params['DataDirectory'])
            fname = str(directory / self.datafile)
        except Exception as e:
            fname = self.datafile = 'None'
            directory = ''

        #--- Which format?
        try:
            self.ts_format = control.params['TS_format']
        except:
            self.ts_format = 'mom'
        if self.ts_format == 'mom':  
            if not fname=='None':
                self.momread(fname)
        elif self.ts_format == 'msf':   
            #--- Are there columns with estimated trajectory models
            try:
                self.use_residuals = control.params['UseResiduals']
            except:
                self.use_residuals = False
            #--- Which column
            try:
                self.observation_column = control.params['ObservationColumn']
            except Exception as e:
                print(e)
                sys.exit()
            if not fname=='None':
                self.msfread(fname)
        else:
            print('Unknown format: {0:s}'.format(self.ts_format))
            sys.exit()

        #--- Inform the user
        if self.verbose==True:
            print("\nFilename                   : {0:s}".format(fname))
            print("TS_format                  : {0:s}".format(self.ts_format))
            print("ScaleFactor                : {0:f}".\
                                                    format(self.scale_factor))
            if self.ts_format == 'msf':   
                print("Observation Column         : {0:d}".\
                                             format(self.observation_column))
                print("Use Residuals              : {0:}".\
                                                    format(self.use_residuals))
            if not fname=='None':
                print("Number of observations+gaps: {0:d}".format(self.m))
                print("Percentage of gaps         : {0:5.1f}".\
					                           format(self.percentage_gaps))



    def create_dataframe_and_F(self,t,obs,mod,period):
        """ Convert np.arrays into Panda DataFrame and create matrix F

        Args:
            t (np.array): array with MJD or sod
            obs (np.array): array with observations
            mod (np.array): array with modelled values
            period (floatt): sampling period (unit is days or seconds)
        """
        
        #--- Store sampling period in this class
        self.sampling_period = period

        #---- Create pandas DataFrame
        self.data = pd.DataFrame({'obs':np.asarray(obs)}, \
                                              index=np.asarray(t))
        if len(mod)>0:
            self.data['mod']=np.asarray(mod)
            
        #--- Create special missing data matrix F
        self.m = len(self.data.index)
        n = self.data['obs'].isna().sum()
        self.F = np.zeros((self.m,n))
        j=0
        for i in range(0,self.m):
            if np.isnan(self.data.iloc[i,0])==True:
                self.F[i,j]=1.0
                j += 1

        #--- Compute percentage of gaps
        self.percentage_gaps = 100.0 * float(n) /float(self.m)



    def momread(self,fname):
        """Read mom-file fname and store the data into the mom class
        
        Args:
            fname (string) : name of file that will be read
        """
        #--- Constants
        TINY = 1.0e-6

        #--- Check if file exists
        if os.path.isfile(fname)==False:
            print('File {0:s} does not exist'.format(fname))
            sys.exit()
        
        #--- Read the file (header + time series)
        t = []
        obs = []
        mod = []
        mjd_old = 0.0
        with open(fname,'r') as fp:
            for line in fp:
                cols = line.split()
                if line.startswith('#')==True:
                    if len(cols)>3:
                        if cols[1]=='sampling' and cols[2]=='period':
                            self.sampling_period = float(cols[3])
                    if len(cols)>2:
                        if cols[0]=='#' and cols[1]=='offset':
                            self.offsets.append(float(cols[2]))
                        elif cols[0]=='#' and cols[1]=='exp':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.postseismicexp.append([mjd,T])
                        elif cols[0]=='#' and cols[1]=='log':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.postseismiclog.append([mjd,T])
                        elif cols[0]=='#' and cols[1]=='tanh':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.ssetanh.append([mjd,T])
                else:
                    if len(cols)<2 or len(cols)>3:
                        print('Found illegal row: {0:s}'.format(line))
                        sys.exit()
                    
                    mjd = float(cols[0])
                    #--- Fill gaps with NaN's
                    if mjd_old>0.0:
                        while abs(mjd-mjd_old-self.sampling_period)>TINY:
                            mjd_old += self.sampling_period
                            t.append(mjd_old)
                            obs.append(np.NAN)
                            if len(cols)==3:
                                mod.append(float(np.NAN))
                            if mjd_old>mjd-TINY:
                                print('Someting is very wrong here....')
                                print('mjd={0:f}'.format(mjd))
                                sys.exit()
                    t.append(mjd)
                    mjd_old = mjd
                    obs.append(self.scale_factor * float(cols[1]))
                    if len(cols)==3:
                        mod.append(self.scale_factor * float(cols[2]))
        
        self.create_dataframe_and_F(t,obs,mod,self.sampling_period)



    def msfread(self,fname):
        """Read msf-file fname and store the data into the Observation class
        
        Args:
            fname (string) : name of file that will be read
        """
        #--- Constants
        TINY = 1.0e-6

        #--- Check if file exists
        if os.path.isfile(fname)==False:
            print('File {0:s} does not exist'.format(fname))
            sys.exit()
        
        #--- Read the file (header + time series)
        t = []
        obs = []
        mjd0 = 0.0
        tt_old = 0.0
        inside_observations = False
        inside_models = False
        inside_offsets = False
        self.model_column = 0
        self.fname_in = fname
        offset_lines = []
        with open(fname,'r') as fp:
            for line in fp:
                cols = line.split()
                if line.startswith('#')==True:
                    if len(cols)>2:
                        if cols[1]=='sampling' and cols[2]=='period':
                            self.sampling_period = float(cols[3])
                    if len(cols)>1:
                        if cols[1]=='Observations':
                            inside_observations = True
                        elif cols[1]=='Models':
                            inside_observations = False
                            inside_models = True
                        elif cols[1]=='Offsets':
                            inside_models = False
                            inside_offset = True

                        #-- Does the line contain column definition?
                        m = re.match('#\s+(\d+)\.\s+(.+)',line)
                        if m:
                            column = int(m.group(1))
                            if inside_observations==True and \
                                              column==self.observation_column:
                                self.column_name = m.group(2)
                            elif inside_models==True and \
                                              m.group(2)==self.column_name:
                                self.model_column = column
                            elif inside_offsets==True and \
                                              column==self.observation_column:
                                offset_lines.append(m.group(2))
 
                else:
                    #--- sanity check
                    if self.use_residuals==True:
                        if self.model_column==0:
                            print('Could not find requested model column')
                            sys.exit()
                    if len(cols)<3:
                        print('Found illegal row: {0:s}'.format(line))
                        sys.exit()
                   
                    mjd = float(cols[0])
                    if mjd0==0.0:
                        mjd0 = mjd
                    sod = float(cols[1])
                    tt = 86400.0*(mjd-mjd0) + sod
                    #--- Fill gaps with NaN's
                    if tt_old>0.0:
                        while abs(tt-tt_old-self.sampling_period)>TINY:
                            tt_old += self.sampling_period
                            t.append(tt_old)
                            obs.append(np.NAN)
                            if tt_old>tt-TINY:
                                print('Someting is very wrong here....')
                                print('mjd={0:f}, sod={1:f}'.format(mjd,sod))
                                sys.exit()
                    t.append(tt)
                    tt_old = tt
                    if cols[self.observation_column-1]=='NaN':
                        obs.append(np.NAN)
                    else:
                        if self.use_residuals==True:
                            y = float(cols[self.observation_column-1])-\
                                       float(cols[self.model_column-1])
                        else:
                            y = float(cols[self.observation_column-1])
                        obs.append(self.scale_factor * y)

        #--- Now that we know MJD0, store offset times
        for line in offset_lines:
            cols = line.split()
            tt = 86400.0*(float(cols[0]) - mjd0) + float(cols[1])
            self.offsets.append(tt)

        #---- Create pandas DataFrame
        self.create_dataframe_and_F(t,obs,[],self.sampling_period)
        
        
        
    def momwrite(self,fname):
        """Write the momdata to a file called fname
        
        Args:
            fname (string) : name of file that will be written
        """
        #--- Try to open the file for writing
        try:
            fp = open(fname,'w') 
        except IOError: 
           print('Error: File {0:s} cannot be opened for written.'. \
                                                         format(fname))
           sys.exit()
        if self.verbose==True:
            print('--> {0:s}'.format(fname))
        
        #--- Write header
        fp.write('# sampling period {0:f}\n'.format(self.sampling_period))
                
        #--- Write header offsets
        for i in range(0,len(self.offsets)):
            fp.write('# offset {0:10.4f}\n'.format(self.offsets[i]))
        #--- Write header exponential decay after seismic event
        for i in range(0,len(self.postseismicexp)):
            [mjd,T] = self.postseismicexp[i]
            fp.write('# exp {0:10.4f} {1:5.1f}\n'.format(mjd,T))
        #--- Write header logarithmic decay after seismic event
        for i in range(0,len(self.postseismiclog)):
            [mjd,T] = self.postseismiclog[i]
            fp.write('# exp {0:10.4f} {1:5.1f}\n'.format(mjd,T))
        #--- Write header slow slip event
        for i in range(0,len(self.ssetanh)):
            [mjd,T] = self.sshtanh[i]
            fp.write('# tanh {0:10.4f} {1:5.1f}\n'.format(mjd,T))
 
        #--- Write time series
        for i in range(0,len(self.data.index)):
            if not math.isnan(self.data.iloc[i,0])==True:
                fp.write('{0:12.6f} {1:13.6f}'.format(self.data.index[i],\
                                                  self.data.iloc[i,0]))
                if len(self.data.columns)==2:
                    fp.write(' {0:13.6f}\n'.format(self.data.iloc[i,1]))
                else:
                    fp.write('\n')
            
        fp.close()
        
       

    def msfwrite(self,fname):
        """Write the msf data to a file called fname
        
        Args:
            fname (string) : name of file that will be written

        self.observation_column, self.column_name are assumed to be known
        after msfread call. If not, then self.fname_in is also '' and a new
        header will be created, setting these two variables

        """

        #--- Constant
        TINY = 1.0e-8

        #--- Initial value, to be changed later
        mjd0 = 0.0

        #--- Create header from scratch
        if self.datafile=="None":
            self.column_name='unknown'
            self.observation_columnn = 3

            lines = []
            lines.append('# sampling period {0:f}\n#\n'.\
                                                format(self.sampling_period))
            lines.append('# Observations\n# ------------\n')
            lines.append('# 1. MJD - Modified Julian Date\n')
            lines.append('# 2. sod - seconds of day\n')
            lines.append('# 3. {0:s}\n#\n'.format(self.column_name))

            lines.append('# Models\n# ------------\n#\n')

            #--- Write offsets in header
            mjd0 = 58849.0
            lines.append('# Offsets\n# ------------\n')
            for i in range(0,len(self.offsets)):
                sod = self.offsets[i]
                mjd = mjd0
                while sod>86400.0:
                    mjd += 1
                    sod -= 86400.0
                lines.append('# 3. {0:6.0f} {1:10.4f}\n'.format(mjd,sod))
 
            lines.append('#\n#=====================================' + \
                                '==========================================\n')

            #--- Add MJD, sod & observation lines
            for i in range(0,self.m):
                mjd = mjd0
                sod = self.data.index[i]
                while sod>86400.0:
                    mjd += 1
                    sod -= 86400.0
                lines.append('{0:6.0f} {1:10.4f} {2:12.6f}\n'.\
                                          format(mjd,sod,self.data.iloc[i,0]))

        else:
            with open(self.fname_in,'r') as fp:
                lines = fp.readlines()

        #--- Try to open the file for writing
        try:
            fp = open(fname,'w')
        except IOError:
           print('Error: File {0:s} cannot be opened for written.'. \
                                                         format(fname))
           sys.exit()

        if self.verbose==True:
            print('--> {0:s}'.format(fname))

        #--- Case of adding anther column 
        inside_observations = False
        inside_models = False
        inside_offsets = False
        i = 0

        for line in lines:

            #--- header stuff
            if line.startswith('#'):
                #--- If new trajectory model is estimated, add name in header
                if len(self.data.columns)==2:
                    if inside_observations==True:
                        #-- Does the line contain column definition?
                        m = re.match('# (\d+)\.\s+(.+)',line)
                        if m:
                            column = int(m.group(1))
                    elif inside_models==True:
                        #-- Does the line contain column definition?
                        m = re.match('# (\d+)\.\s+(.+)',line)
                        if m:
                            column = int(m.group(1))
                            if m.group(2)==self.column_name:
                                self.model_column = column
                        else:
                            if not line.startswith('# ------------') and \
                                                        self.model_column==0:
                                self.model_column = column+1
                                fp.write('# {0:d}. {1:s}\n'.format(column+1,
                                                            self.column_name))
                                inside_models=False

                    #--- Which block are we processing?
                    cols = line.split()
                    if len(cols)>1:
                        if cols[1]=='Observations':
                            inside_observations = True
                        if cols[1]=='Models':
                            inside_observations = False
                            inside_models = True
                        elif cols[1]=='Offsets':
                            inside_models = False
 
                #--- Just copy header line to new file
                fp.write(line)
            else:
                #--- Get time to see if this is equal to data.index[i]
                cols = line.split()
                mjd = float(cols[0])
                sod = float(cols[1])
                if mjd0==0.0:
                    mjd0 = mjd

                tt = 86400.0*(mjd - mjd0) + sod
                while self.data.index[i]<tt:
                    i += 1
                if abs(tt - self.data.index[i])>TINY:
                    print('Expected regularly spaced data!')
                    print('file time = {0:f}, array time = {1:f}'.\
                                            format(tt,self.data.index[i]))
                    sys.exit()

                #--- Replace observation values
                cols = line.split()
                fp.write('{0:>6s} {1:>10s}'.format(cols[0],cols[1]))
                for j in range(2,self.observation_column-1):
                    fp.write('{0:>12s} '.format(cols[j]))
                if math.isnan(self.data.iloc[i,0])==True:
                    fp.write('{0:>12s}'.format('NaN'))
                else:
                    fp.write('{0:12.6f}'.format(self.data.iloc[i,0]))

                if len(self.data.columns)==2:
                    #--- Replace Model values
                    for j in range(self.observation_column,\
                                                         self.model_column-1):
                        fp.write(' {0:>12s}'.format(cols[j]))
                    if math.isnan(self.data.iloc[i,0])==True:
                        fp.write('{0:>12s}'.format('NaN'))
                    else:
                        fp.write('{0:12.6f}'.format(self.data.iloc[i,1]))
                    for j in range(self.model_column,len(cols)):
                        fp.write(' {0:>12s}'.format(cols[j]))
                else:
                    for j in range(self.observation_column,len(cols)):
                        fp.write(' {0:>12s}'.format(cols[j]))
                fp.write('\n')
                
                #--- Go to next item in array
                i += 1

        fp.close()



    def show_results(self,output):
        """ add info to json-ouput dict
        """
        
        output['N'] = self.m
        output['gap_percentage'] = self.percentage_gaps 
        if self.ts_format=='mom':
            output['TimeUnit'] = 'days'
        else:
            output['TimeUnit'] = 'seconds'



    def add_offset(self,t):
        """ Add time t to list of offsets
        
        Args:
            t (float): modified julian date or second of day of offset
        """

        EPS   = 1.0e-6
        found = False
        i     = 0
        while i<len(self.offsets) and found==False:
            if abs(self.offsets[i]-t)<EPS:
                found = True
            i += 1
        if found==False:
            self.offsets.append(t)



    def set_NaN(self,index):
        """ Set observation at index to NaN and update matrix F

        Args:
            index (int): index of array which needs to be set to NaN
        """

        self.data.iloc[index,0] = np.nan 
        dummy = np.zeros(self.m)
        dummy[index] = 1.0
        self.F = np.c_[ self.F, dummy ] # add another column to F



    def add_mod(self,xhat):
        """ Add estimated model as column in DataFrame

        Args:
            xhat (array float) : estimated model
        """

        self.data['mod']=np.asarray(xhat)



    def write(self,fname):
        """ Select correct subroutine for writing to file

        Args:
            fname (string): complete name of file
        """

        if self.ts_format=='mom':
            self.momwrite(fname)
        elif self.ts_format=='msf':
            self.msfwrite(fname)
        else:
            print('unknown ts_format: {0:s}'.format(self.ts_format))
            sys.exit()



# -*- coding: utf-8 -*-
#
# Simple class providing the first order autoregressive noise model
#
# This file is part of Hector.
#
# Hector is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# Hector is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# Hector. If not, see <https://www.gnu.org/licenses/>.
#
# 18/8/2022 Machiel Bos, Santa Clara
#==============================================================================

import numpy as np
import math
from hectorp.control import Control

#==============================================================================
# Subroutines
#==============================================================================


class AR1:

    def __init__(self):
        """ initialise class
        """

        #--- Get instances of classes
        control = Control()

        #--- Check if phi is given in control file
        try:
            self.phi_fixed = control.params['phi_ar1_fixed']
        except:
            self.phi_fixed = math.nan



    def get_Nparam(self):
        """ Return the number of parameters in White noise model
        
        Returns
        -------
        self.Nparam (int) : total number of parameters === 1 - phi
        """

        return 1


        
    def create_t(self,m,k,param):
        """ Create first row of covariance matrix of white noise
    
        Arguments
        ---------
        m (int) : length of time series
        k (int) : index of param
        param (array float) : phi
        
        Returns
        -------
        t (row (m,1)) : first row Toeplitz covariance matrix 
        k_new (int)   : shifted index in param array
        """

        #--- Parse param
        phi = param[k]
        k_new = k+1   # increase k for next model

        #--- Create first row vector of Covariance matrix
        t = np.zeros(m)

        #--- first, take care of power of phi
        t[0] = 1.0/(1.0 - phi*phi)
        for i in range(1,m):
            t[i] = t[i-1]*phi
        
        return t, k_new


  
    def penalty(self,k,param):
        """ Computes penalty for varying annual noise

        Arguments
        ---------
        k (int) : index of param
        param (array float) : phi
        
        Returns
        -------
        penalty (float)
        """

        penalty = 0.0
        if math.isnan(self.phi_fixed)==True:
            LARGE = 1.0e8
            phi = param[k]
            #--- Check range of parameters
            if phi<0.0:
                penalty += (0.0 - phi)*LARGE
                param[k] = 0.0
            elif phi>0.99999:
                penalty += (phi - 0.99999)*LARGE
                param[k] = 0.99999
         
        return penalty



    def show_results(self,output_single,k,noise_params,sigma):
        """ show estimated noiseparameters

        Args:
            output_single (dictionary) : where values for json file are saved
            k (int) : index where we should start reading noise_params
            noise_params (float-array) : fractions + noise model parameters
            sigma (float) : noise amplitude of white noise
        """


        #--- Get some info from other classes
        control = Control()
        try:
            verbose = control.params['Verbose']
        except:
            verbose = True

        k_new = k
        if verbose==True:
            unit = control.params['PhysicalUnit']
            print('sigma     = {0:7.4f} {1:s}'.format(sigma,unit))
            if math.isnan(self.phi_fixed)==True:
                phi = noise_params[k]
                print('phi       = {0:7.4f}'.format(phi))
            else:
                output_single['phi'] = self.phi_fixed
                print('phi       = {0:7.4f} (fixed)'.format(self.phi_fixed))

        output_single['sigma'] = sigma
        if math.isnan(self.phi_fixed)==True:
           phi = noise_params[k]
           output_single['phi'] = phi
           k_new = k+1
        else:
           output_single['phi'] = self.phi_fixed

        return k_new




# -*- coding: utf-8 -*-
#
# This program creates synthetic noise.
#
#  This script is part of HectorP 0.1.1
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 6/2/2022 Machiel Bos, Santa Clara
#===============================================================================

import os
import math
import time
import sys
import numpy as np
import argparse
from hectorp.control import Control
from hectorp.observations import Observations
from scipy import signal
from pathlib import Path
from scipy.special import kv

#===============================================================================
# Subroutines
#===============================================================================

def mactrick(m,gamma_x):
    """ Use Cholesky decomposition to get first column of C

    Args:
        m (int) : length of time series
        gamma_x (float array) : first column of covariance matrix

    Returns:
        h : array of float with impulse response
    """

    U = np.zeros((m,2))  # C = U’*U
    V = np.zeros((m,2))
    h = np.zeros(m)

    #--- define the generators u and v 
    U[:,0] = gamma_x/math.sqrt(gamma_x[0]) 
    V[1:m,0] = U[1:m,0]
    h[m-1] = U[m-1,0]

    k_old =0;
    k_new =1;
    for k in range(0,m-1):
        sin_theta = V[k+1,k_old]/U[k,k_old]
        cos_theta = math.sqrt(1.0-pow(sin_theta,2))
        U[k+1:m,k_new] = ( U[k:m-1,k_old] - sin_theta*V[k+1:m,k_old])/cos_theta
        V[k+1:m,k_new] = (-sin_theta*U[k:m-1,k_old] + V[k+1:m,k_old])/cos_theta
        h[m-1-k] = U[m-1,k_new]
  
        k_old = 1-k_old
        k_new = 1-k_new

    return h



def create_h(m,noisemodel,dt,ts_format,control):
    """ Create impulse function

    Args:
        m (int) : length of time series
        noisemodel (string) : abreviation of noise model (PL, WL, GGM)
        dt (float) : sampling period in days
        ts_format (string) : 'mom' or 'msf'

    Returns:
        [sigma,h] : noise amplitude + array of float with impulse response
    """

    #--- Constant, small number
    EPS = 1.0e-8
    threshold = 100.0

    #--- Array to fill impulse response
    h = np.zeros(m)

    #--- Ask noise parameter values
    if noisemodel=='White':
        print('white noise amplitude: ', end='')
        sigma = float(input())
        h[0] = sigma
    elif noisemodel in ['Powerlaw','Flicker','RandomWalk']:
        if noisemodel=='Powerlaw':
            print('spectral index kappa: ', end='')
            kappa = float(input())
            if kappa<-2.0-EPS or kappa>2.0+EPS:
                print('kappa shoud lie between -2 and 2 : {0:f}'.format(kappa))
                sys.exit()
            d = -kappa/2.0
            print('power-law noise amplitude: ', end='')
        elif noisemodel=='Flicker':
            d = 0.5
            print('flicker noise amplitude: ', end='')
        else:
            d = 1.0
            print('random walk noise amplitude: ', end='')
        sigma = float(input())
        if ts_format=='mom':
            sigma *= math.pow(dt/365.25,0.5*d)  # Already adjust for scaling
        elif ts_format=='msf':
            sigma *= math.pow(dt/3600.0,0.5*d)  # Already adjust for scaling
        else:
            print('unknown scaling: {0:s}'.format(ts_format))
            sys.exit()

        h[0] = 1.0;
        for i in range(1,m):
            h[i] = (d+i-1.0)/i * h[i-1]
    elif noisemodel=='GGM':
        try:
            phi = control.params['GGM_1mphi']
        except:
            print('factor 1-phi (to avoid 0.999999...): ', end='')
            phi = float(input())
            if phi<0.0 or phi>1.0+EPS:
                print('1-phi should lie between 0 and 1: {0:f}'.format(phi))
                sys.exit()
        print('spectral index kappa: ', end='')
        kappa = float(input())
        if kappa<-2.0-EPS or kappa>2.0+EPS:
            print('kappa shoud lie between -2 and 2 : {0:f}'.format(kappa))
            sys.exit()
        d = -kappa/2.0
        print('power-law noise amplitude: ', end='')
        sigma = float(input())
        if ts_format=='mom':
            sigma *= math.pow(dt/365.25,0.5*d)  # Already adjust for scaling
        elif ts_format=='msf':
            sigma *= math.pow(dt/3600.0,0.5*d)  # Already adjust for scaling
        else:
            print('unknown scaling: {0:s}'.format(ts_format))
            sys.exit()
        h[0] = 1.0;
        for i in range(1,m):
            h[i] = (d+i-1.0)/i * h[i-1] * (1.0-phi)
    elif noisemodel=='VaryingAnnual':
        if not ts_format=='mom':
            print('Please think again, ts_format is not mom!!!')
            sys.exit()
        try:
            phi = control.params['phi_fixed']
        except:
            print('phi: ', end='')
            phi = float(input())
        print('varying-annual noise amplitude: ', end='')
        sigma = float(input())

        #--- create gamma_x
        t = np.zeros(m)
        t[0] = 1.0/(2.0 * (1.0 - phi*phi))
        for i in range(1,m):
            t[i] = t[i-1]*phi

        #--- next, multiply with cosine dt = in days
        for i in range(1,m):
            t[i] *= math.cos(2*math.pi * i * dt/365.25)

        h = mactrick(m,t)

    elif noisemodel=='Matern':
        try:
            lamba = control.params['lambda_fixed']
        except:
            print('lambda: ', end='')
            lamba = float(input())
        try:
            d = -0.5*control.params['kappa_fixed']
        except:
            print('kappa: ', end='')
            d = -0.5*float(input())
        print('Matern noise amplitude: ', end='')
        sigma = float(input())

        #--- create gamma_x
        t = np.zeros(m)

        #--- Constant
        alpha = 2.0*d
        c0 = 2.0/(math.gamma(alpha-0.5) * pow(2.0, alpha-0.5))

        #--- Use Modified Bessel Function, second order
        t[0] = 1.0 #-- check Eq. (62) and set tau=0
        i = 1
        tau = float(i)
        while i<m:
            if tau>threshold/lamba:
                #--- Eq. (61)
                t[i] = c0*math.sqrt(math.pi/2.0)*pow(lamba*tau,alpha-0.5)\
                                                        * math.exp(-lamba*tau)
            else:
                #--- Eq. (60)
                t[i] = c0*pow(lamba*tau,alpha-0.5) * kv(alpha-0.5,lamba*tau)

            #--- Next tau
            i += 1
            tau = float(i)

        h = mactrick(m,t)

    elif noisemodel=='AR1':
        print('phi: ', end='')
        phi = float(input())
        print('AR1 noise amplitude: ', end='')
        sigma = float(input())

        #--- Create first row vector of Covariance matrix
        t = np.zeros(m)

        #--- first, take care of power of phi
        t[0] = 1.0/(1.0 - phi*phi)
        for i in range(1,m):
            t[i] = t[i-1]*phi

        h = mactrick(m,t)

    else:
        print('Unknown noisemodel: {0:s}'.format(noisemodel))
        sys.exit()


    return [sigma,h]


#===============================================================================
# Main program
#===============================================================================

def main():

    print("\n***************************************")
    print("    simulatenoise, version 0.1.1")
    print("***************************************")

    #--- Parse command line arguments in a bit more professional way
    parser = argparse.ArgumentParser(description= 'Simulate noise time series')

    #--- List arguments that can be given 
    parser.add_argument('-i', required=False, default='simulatenoise.ctl', \
                                      dest='fname', help='Name of control file')

    args = parser.parse_args()

    #--- parse command-line arguments
    fname = args.fname

    #--- Read control parameters into dictionary (singleton class)
    control = Control(fname)
    observations = Observations()

    #--- Some variables that define the runs
    directory     = Path(control.params['SimulationDir'])
    label         = control.params["SimulationLabel"]
    n_simulations = control.params["NumberOfSimulations"]
    m             = control.params["NumberOfPoints"]
    dt            = control.params["SamplingPeriod"]
    ms            = control.params["TimeNoiseStart"]
    noisemodels   = control.params['NoiseModels']
    try:
        repeatablenoise = control.params['RepeatableNoise']
    except:
        repeatablenoise = False
    try:
        missingdata = control.params['MissingData']
        perc_missingdata = control.params['PercMissingData']
    except:
        missingdata = False
        perc_missingdata = 0.0
    try:
        includeoffsets = control.params['Offsets']
    except:
        includeoffsets = False
    try:
        trend = control.params['Trend']
        nominalbias = control.params['NominalBias']
    except:
        trend = 0.0
        nominalbias = 0.0
    try:
        annualsignal = control.params['AnnualSignal']
    except:
        annualsignal = 0.0


    #--- Start the clock!
    start_time = time.time()

    #--- Already create all the impulse functions
    if isinstance(noisemodels,list)==False:
        noisemodels = [noisemodels]
    n_models = len(noisemodels)
    zeros    = np.zeros(m+ms)
    h        = [zeros]*n_models
    sigma    = [0.0]*n_models
    j = 0
    for noisemodel in noisemodels:
        [sigma[j],h[j]] = create_h(m+ms,noisemodel,dt,observations.ts_format,\
                                                                    control)
        j += 1

    #--- Already create time array
    if observations.datafile=='None' and observations.ts_format=='mom':
        t = np.zeros(m);
        t[0] = 51544.0 # 1 January 2000
        for i in range(1,m):
            t[i] = t[i-1] + dt
    elif observations.datafile=='None' and observations.ts_format=='msf':
        t = np.zeros(m);
        for i in range(1,m):
            t[i] = t[i-1] + dt
    else:
        print('problem, not implemented yet....')
        sys.exit()

    #--- Create random number generator
    if repeatablenoise==True:
        rng = np.random.default_rng(0)
    else:
        rng = np.random.default_rng()

    #--- Does the directory exists?
    if not os.path.exists(directory):
       os.makedirs(directory)

    #--- Run all simulations
    for k in range(0,n_simulations):

        #--- Open file to store time-series
        datafile = label + '_' + str(k) + "." + observations.ts_format
        fname = str(directory.resolve()) + '/' + datafile

        #--- Create the synthetic noise
        y = np.zeros(m)

        #--- Create trend + annual signal
        for j in range(0,m):
            tt = t[j]-t[0]
            y[j] += nominalbias + trend*tt/365.25 + \
                                     annualsignal*math.cos(2*math.pi*tt/365.25)

        for j in range(0,len(sigma)):
            w = sigma[j] * rng.standard_normal(m+ms)
            y += signal.fftconvolve(h[j], w)[0:m]

        #--- convert this into Panda dataframe
        observations.create_dataframe_and_F(t,y,[],dt)

        #--- write results to file
        observations.write(fname)


    #--- Show time lapsed
    print("--- {0:8.3f} seconds ---\n".format(float(time.time() - start_time)))
